/* 
 * tclMacSock.c
 *
 *	Channel drivers for Macintosh sockets.
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclMacSock.c 1.3 96/02/21 10:33:44
 */

#include "tclInt.h"
#include "tclPort.h"
#include <Aliases.h>
#include <Errors.h>
#include <Files.h>
#include <Gestalt.h>
#include <Processes.h>
#include <Strings.h>

/*
 * Static routines for this file:
 */

static int		TcpBlockMode _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, int mode));
static int		TcpBufferSize _ANSI_ARGS_((
    			    ClientData instanceData));
static int		TcpInput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, char *buf, int toRead,
			    int *errorCode));
static int		TcpOutput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File outFile, char *buf, int toWrite,
			    int *errorCode));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));

/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",			/* Type name. */
    TcpBlockMode,		/* Set blocking or
                                 * non-blocking mode.*/
    TcpBufferSize,		/* Buffer size. */
    TcpClose,			/* Close proc. */
    TcpInput,			/* Input proc. */
    TcpOutput,			/* Output proc. */
    NULL			/* Seek proc. */
};

/*
 * This is the size of the channel name for File based channels
 */

#define CHANNEL_NAME_SIZE	64
static char channelName[CHANNEL_NAME_SIZE+1];

/*
 * The preferred buffer size for Unix channels.
 */

#define CHANNEL_BUF_SIZE	4096

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    Tcl_File sock;			/* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;		/* The data for the accept proc. */
} TcpState;

/*
 * Globals for holding information about OS support for sockets.
 */
static int socketsTestInited = false;
static int hasSockets = false;

/*
 *----------------------------------------------------------------------
 *
 * TcpBufferSize --
 *
 *	Returns the preferred buffer size for unix IO channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return CHANNEL_BUF_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    int written;
    int fd;

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(outFile, NULL);
    written = write(fd, buf, (size_t) toWrite);
    if (written > -1) {
        return written;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

int
TcpInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    struct stat statBuf;
    int fd;
    int bytesRead;			/* How many bytes were read? */

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(inFile, NULL);
    if (fstat(fd, &statBuf) != 0) {
        *errorCode = errno;
        return -1;
    }
    if ((statBuf.st_size > 0) && (statBuf.st_size < bufSize)) {
        bufSize = statBuf.st_size;
    }
    bytesRead = read(fd, buf, (size_t) bufSize);
    if (bytesRead > -1) {
        return bytesRead;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpBlockMode --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    int curStatus;
    int fd;

    if (inFile != NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }
    if (outFile != NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    TcpState *statePtr = (TcpState *) instanceData;
    Tcl_File sock = statePtr->sock;
    int errorCode = 0;

    /*
     * We assume that inFile==outFile==sock and so only clean up one of them.
     */
    
    if (close((int) Tcl_GetFileInfo(sock, NULL)) < 0) {
	errorCode = errno;
    }
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sock);

    ckfree((char *) statePtr);

    Tcl_FreeFile(sock);
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed. On failure, the routine also
 *	sets the output argument errorCodePtr to the error code.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
{
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    Tcl_File handle;			/* Socket file handle. */
    int sock;				/* Socket file descriptor */
    int status;
    Tcl_Channel chan;
    TcpState *sockState;

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = (short) (port & 0xFFFF);
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
            Tcl_SetErrno(errno);
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "couldn't open socket: ",
                        Tcl_PosixError(interp), (char *) NULL);
            }
	    return (Tcl_Channel) NULL;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
	return (Tcl_Channel) NULL;
    }

    /*
     * Make server socket close gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));
        
    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, CHANNEL_BUF_SIZE);

    handle = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    sockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    sockState->sock = handle;
    sockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
    sockState->acceptProcData = (ClientData) NULL;

    /*
     * Attempt to connect. The connect may fail at present with an
     * EINPROGRESS but at a later time it will complete. The caller
     * will set up a file handler on the socket if she is interested in
     * being informed when the connect completes.
     */
    
    status = connect(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr));
    if (status < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        Tcl_FreeFile(handle);
        ckfree((char *) sockState);
        return (Tcl_Channel) NULL;
    }

    sprintf(channelName, "sock%d", (int) sock);
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, handle, handle,
            (ClientData) sockState);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        Tcl_FreeFile(handle);
        ckfree((char *) sockState);
        return NULL;
    } else {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    Tcl_File handle;			/* Socket file handle. */
    int sock;				/* Socket file descriptor */
    int status;
    TcpState *sockState;
    Tcl_Channel chan;

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = (short) (port & 0xFFFF);
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
            Tcl_SetErrno(errno);
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "couldn't open socket: ",
                        Tcl_PosixError(interp), (char *) NULL);
            }
            return (Tcl_Channel) NULL;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        return (Tcl_Channel) NULL;
    }

    /*
     * Set up to reuse server addresses automatically.
     */
    
    status = 1;
    (void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
            sizeof(status));

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, CHANNEL_BUF_SIZE);

    /*
     * Bind to the specified port.
     */
    
    status = bind(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr));
    if (status < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        return (Tcl_Channel) NULL;
    }

    /*
     * Listen for connections.
     */

    listen(sock, 5);

    /*
     * Make server socket close down gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));

    handle = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    sockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    sockState->sock = handle;
    sockState->acceptProc = acceptProc;
    sockState->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(handle, TCL_READABLE, TcpAccept,
            (ClientData) sockState);
    
    sprintf(channelName, "sock%d", (int) sock);
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) sockState);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        ckfree((char *) sockState);
        close(sock);
        Tcl_DeleteFileHandler(handle);
        Tcl_FreeFile(handle);
        return NULL;
    } else {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop,
 *	and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    TcpState *sockState;		/* Client data of server socket. */
    int newsock;			/* The new client socket */
    Tcl_File newFile;			/* Its file. */
    TcpState *newSockState;		/* State for new socket. */
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel  chan;			/* Channel instance created. */

    sockState = (TcpState *) data;

    len = sizeof(struct sockaddr_in);
    newsock = accept((int) Tcl_GetFileInfo(sockState->sock, NULL),
	    (struct sockaddr *)&addr, &len);
    if (newsock < 0) {
        return;
    }
    
    newFile = Tcl_GetFile((ClientData) newsock, TCL_UNIX_FD);
    if (newFile) {
        newSockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));

        newSockState->sock = newFile;
        newSockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
        newSockState->acceptProcData = (ClientData) NULL;
        
        sprintf(channelName, "sock%d", (int) newsock);
        chan = Tcl_CreateChannel(&tcpChannelType, channelName, newFile,
                newFile, (ClientData) newSockState);
        if (chan == (Tcl_Channel) NULL) {
            ckfree((char *) newSockState);
            close(newsock);
            Tcl_FreeFile(newFile);
	} else {
            if (sockState->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
                (sockState->acceptProc) (sockState->acceptProcData, chan,
                        inet_ntoa(addr.sin_addr));
            }
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetHostName --
 *
 *	Returns the name of the local host.
 *
 * Results:
 *	Returns a string containing the host name, or NULL on error.
 *	The returned string must be freed by the caller.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static char hostname[255];	/* This buffer should be big enough for
                                 * hostname plus domain name. */
static int  hostnameInited = 0;

char *
Tcl_GetHostName()
{
    if (hostnameInited) {
        return hostname;
    }
    
    if (TclHasSockets(NULL) != TCL_OK) {
	hostname[0] = '\0';
        hostnameInited = 1;
	return hostname;
    }

    if (gethostname(hostname, 100) == 0) {
        hostnameInited = 1;
        return hostname;
    }
    return (char *) NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TclHasSockets --
 *
 *	This function determines whether sockets are available on the
 *	current system and returns an error in interp if they are not.
 *	Note that interp may be NULL.  This call uses the Macintosh
 *	gestalt function to determine the existance of Mac Tcp.
 *
 * Results:
 *	Returns TCL_OK if the system supports sockets, or TCL_ERROR with
 *	an error in interp.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

#define gestaltMacTCPVersion 'mtcp'
int
TclHasSockets(interp)
    Tcl_Interp *interp;
{
    long response;
    
static int socketsTestInited = false;
static int hasSockets = false;
    if (!socketsTestInited) {
	if (Gestalt(gestaltMacTCPVersion, &response) == noErr) {
	    hasSockets = true;
	} else {
	    hasSockets = false;
	}
	socketsTestInited = true;
    }
    if (hasSockets) {
	return TCL_OK;
    }
    if (interp != NULL) {
	Tcl_AppendResult(interp, "sockets are not available on this system",
		NULL);
    }
    return TCL_ERROR;
}
